#!BPY

"""
Name: 'Vertices to Faces'
Blender: 241
Group: 'Mesh'
Tooltip: 'All selected connected/unconnected vertices are cennected and faces are build.'
"""

__author__ = "Apple Grew"
__email__ = ["Apple Grew, a-p-p-l-e-g-r-e-w@gmail.com (remove the dashes)"]
__version__="0.3"

__bpydoc__ = """\
Usage: Select the vertices of the mesh which u want to link and run this script.
       Click on 'Make quads?' if u want that quads are made wherever possible, else
       move your mouse out of this dialog box (now only triangles will be made).

Notes:
1) This is alpha version of script.
2) This is designed to work on planes (maybe flat or curved).
3) There is a serious limitation right now. The script only 'sees' the values of the x & y coordinates
   not the z coordinate. Hence it is like looking at the mesh along its local z-axis. Hence if u have a
   plane that is vertical along the local z-axis then it will look just like a line to the script. This
   may result in unexpected results.
"""

# --------------------------------------------------------------------------
# Vertices to Faces
# --------------------------------------------------------------------------
# Author: Apple Grew
# Acknowledgement: Rohit Singh for the basic algo of tris to quad.
# --------------------------------------------------------------------------
# ***** BEGIN GPL LICENSE BLOCK *****
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# ***** END GPL LICENCE BLOCK *****
# -------------------------------------------------------------------------- 

from Blender import Object, Draw, Window, NMesh, Mathutils
import math, copy

DEBUG=False
INFINITE='infinite'
USE_EXPERIMENTAL_ALGO=True

def DrawProgressBar(value,text):
    #Window.DrawProgressBar(value,text)
    #print text+' ' +str(value*100)+'%'
    pass
    
def error(msg):
    Draw.PupMenu("Error%t|"+msg)

def calStdDeviation(X,deviationAbout=None): #where X is a List of elements of whose standard
    #deviation is to be found.
    if len(X)==0:
        return 0
    
    if deviationAbout==None:
        avg=0
        for xi in X:
            avg=avg+xi
        avg=avg/len(X)
        deviationAbout=avg
    
    sqSum=0
    for xi in X:
        sqSum=sqSum+ (xi-deviationAbout)**2
    
    return int(math.sqrt(sqSum/len(X))*1000000) #I have observed that after 6th decimal place the std dev always differs even when it should all be equal.

def calDis(v1,v2): #calculates distance between two vertices.
    v1=v1.co #getting coordinate.
    v2=v2.co
    return math.sqrt((v1.x-v2.x)**2+(v1.y-v2.y)**2)

def isLinesIntersecting(v1,v2, v3,v4,vertInLoopMode=False):
    #Double point is a pt. where the pt. of intersection of the two line segments is the either terminal pt.s of the other line segment (i.e. v3,v4)
    x1=v1.co.x
    y1=v1.co.y
    
    x2=v2.co.x
    y2=v2.co.y
    
    x3=v3.co.x
    y3=v3.co.y
    
    x4=v4.co.x
    y4=v4.co.y
    
    if (x1,y1)==(x3,y3) or (x1,y1)==(x4,y4)\
    or (x2,y2)==(x3,y3) or (x2,y2)==(x4,y4):
        if vertInLoopMode:
            return 1
        else:
            return False
    
    if (x2-x1)==0:
        k1=INFINITE
    else:
        k1=(y2-y1)/(x2-x1)
    if (x4-x3)==0:
        k2=INFINITE
    else:
        k2=(y4-y3)/(x4-x3)
    if k1==k2: #the lines are parallel
        #The lines are parallel, well what if they lie on top of each other? Huh?
        #Checking for this case. It is simple to check just put one of the other lines
        #vert (say v3) in eq of line 1 and if it is zero the bingo they lie on each other.
        flag=0
        if k1==INFINITE:
            if x3==x1: flag=1
        else:
            y=y1+k1*(x3-x1)
            if y==y3: flag=1
        
        if flag==1:
            if x1<x2:
                X1,X2=x1,x2
            else:
                X1,X2=x2,x1
            if y1<y2:
                Y1,Y2=y1,y2
            else:
                Y1,Y2=y2,y1
            if (X1<=x3 and x3<=X2 and Y1<=y3 and y3<=Y2)\
            or (X1<=x4 and x4<=X2 and Y1<=y4 and y4<=Y2): flag=2
        
        if flag==2:
            if vertInLoopMode:
                return 2
            else:
                return True
        else:
            if vertInLoopMode:
                return 0
            else:
                return False
    
    if k1==INFINITE:
        x=x1
        y=y3+k2*(x-x3)
    else:
        if k2==INFINITE:
            x=x3
        else:
            x=(y3-y1+k1*x1-k2*x3)/(k1-k2)
        y=y1+k1*(x-x1)
##    if DEBUG: print 'Pt. of Intersection (x,y): '+str((x,y))
    
    if vertInLoopMode and (x,y)==(x3,y3):
        return 3
    if vertInLoopMode and (x,y)==(x4,y4):
        return 4
    
    flag=0
    if x1<x2:
        X1,X2=x1,x2
    else:
        X1,X2=x2,x1
    if y1<y2:
        Y1,Y2=y1,y2
    else:
        Y1,Y2=y2,y1
    if X1<=x and x<=X2 and Y1<=y and y<=Y2:
        flag=1
    
    if x3<x4:
        X1,X2=x3,x4
    else:
        X1,X2=x4,x3
    if y3<y4:
        Y1,Y2=y3,y4
    else:
        Y1,Y2=y4,y3    
    if X1<=x and x<=X2 and Y1<=y and y<=Y2:
        flag+=1
    
    if flag==2:
        if vertInLoopMode:
            return 1
        else:
            return True
    else:
        if vertInLoopMode:
            return 0
        else:
            return False

def isConnectable(vert,vd,doNotConnect,Edges,vi1,vi2):
    #checks wheather vertx v1 can connect with v2. (note: vi1 and vi2 are
    #indices not the obeject itself.
    #2 vertices are not connectable if while connecting them it crosses
    #another edge.
    #edges is a list of all edges present in the object even if none of its verts are selected.

    for dnC in doNotConnect:
        if dnC.count(vert[vi1])!=0 and dnC.count(vert[vi2])!=0:
            return False
    
    for vli in range(len(vd)):
        if len(vd[vli])==0:
            continue
        for i in vd[vli]:
            if isLinesIntersecting(vert[vli],vert[i],vert[vi1],vert[vi2]): return False
    
    #Now checking possible crossings (intersectiont) with the edges in 'Edges'.
    if len(Edges)!=0:
        for e in Edges:
            if not isInSamePlane([e.v1,e.v2,vert[vi1],vert[vi2]]): continue
            if isLinesIntersecting(e.v1,e.v2,vert[vi1],vert[vi2]): return False
    return True

def isInSamePlane(v):#checks wheather all the given vertices lie on the same plane or not.
    v1=v[0].co
    v2=v[1].co
    v3=v[2].co
    v4=v[3].co
    
    #creating matrix which represents the eq of plane.
    matrixPlane=Mathutils.Matrix([v4.x-v1.x, v4.y-v1.y, v4.z-v1.z],\
                                 [v4.x-v2.x, v4.y-v2.y, v4.z-v2.z],\
                                 [v4.x-v3.x, v4.y-v3.y, v4.z-v3.z])
##    if DEBUG:
##        print '.....matrixPlane dump.....'
##        print matrixPlane
##        print '.....dump complete.....'
    if matrixPlane.determinant()==0: #then the vert.s lie in a plane.
        return True
    else:
        return False

def signOf(no):
    if no<0:
        return 1
    else:
        return 0
    
def isTrigulizable(tvl,vert):
    for v in vert:
        x=v.co.x
        y=v.co.y
        SKIP=False
        for i in range(len(tvl)):
            if (x,y)==(tvl[i].co.x,tvl[i].co.y):
                SKIP=True
                break
        if SKIP: continue
        
        outCount=0
        for i in range(len(tvl)):
            if i==2:
                j=0
            else:
                j=i+1
            if j==2:
                k=0
            else:
                k=j+1
            x1=tvl[i].co.x
            y1=tvl[i].co.y
            
            x2=tvl[j].co.x
            y2=tvl[j].co.y
            
            x3=tvl[k].co.x
            y3=tvl[k].co.y
            
            if (x2-x1)==0:
                if signOf((x1-x3))!=signOf((x1-x)):
                    outCount+=1
                continue
            
            k=(y2-y1)/(x2-x1)
            eq1=y1-y+k*(x-x1)
            eq2=y1-y3+k*(x3-x1)
            if signOf(eq1)!=signOf(eq2):
                outCount+=1
        if outCount==0:
            return False
    return True

def DetectLoops(vd):
    loops=[]
    
    #Identifying vertex loops.
    adj=[]#it is the adjacency list
    for a in vd:
        adj.append([])
    status=[] #status list
    for i in range(len(vd)):
        status.append(0)
        for j in range(len(vd[i])):
            adj[i].append(vd[i][j])
            adj[vd[i][j]].append(i)
    
    for I in range(len(adj)):
        stack=[I]
        loop=[]
        while len(stack)!=0:
            i=stack.pop()
            if status[i]==1:
                if i==I and len(loop)>2:
                    T=[]
                    for t in loop: #copying loop into T. This IS VERY IMPORTANT.
                        T.append(t)
                    loops.append(T)
##                        if DEBUG: print '____loops____ '+str(loops)
                    T=[]
                    #The copying of loop into T is important since if we just append loop to loops then
                    #loops & loop both point to the SAME object and now in the follwing step when loop is
                    #poped then data from that object (to which loops also points) is removed, hence content
                    #of loops also changes, which is undesired. It took me hours to figure out. Phew!
                    
                    loo=[]
                    loo.extend(loop)
                    for l in range(len(loo)):
                        breakout=False
                        for m in adj[loo[len(loo)-1-l]]:
                            if len(loop)==0:
                                breakout=True
                                break
                            if status[m]==1: loop.pop()
                        if breakout: break
                    loo=[]
                continue
            
            loop.append(i)
            status[i]=1
            stack.extend(adj[i])
##                if DEBUG:
##                    print '$$$$$o loop o$$$$$ '+str(loop)
##                    print
    if DEBUG:
        print '[][][]Loops dump[][][]'
        for loo in loops:
            print loo
        print '[][][]dump complete[][][]'
    return loops

def RankLoops(loops,vert):
    rank=[]
    if len(loops)!=0:
        #Ranking loops.
        rank=[0]
        for r in range(1,len(loops)):
            rank.append([])
        for i in range(1,len(loops)):
            for j in range(0,i):
                flag=0
                continue_j_loop=False
                for vi in loops[i]:
                    f=isInLoop(loops[j],vert,vi)
##                        print 'loop (i,j)'+str((i,j))+'---flag value--->>> ' +str(f)
                    if f==0: #current vertex lie on j
                        f=flag
                    if f==1: #current vertex lie inside j
                        if flag==-1: #previous vertex was outside j
                            rank[i]=rank[j]
                            continue_j_loop=True
                            break
                    if f==-1: #current vertex lie outside j
                        if flag==1: #previous vertex was inside j
                            rank[i]=rank[j]
                            continue_j_loop=True
                            break
                    flag=f
                if continue_j_loop: continue
                if flag==1: #all verts of i lie within j or on
                    r=rank[j]+1
                    if rank[i]==[] or r>rank[i]:
                        rank[i]=r
                if flag==-1: #all verts of i lie outside j or on
                    r=rank[j]-1
                    if rank[i]==[] or r<rank[i]:
                        rank[i]=r
    return rank
    
def isInLoop(loop,vert,v1,v2=None,ignoreThatVertexLieOnTheLoop=False):
    class CO:
        x=0
        y=0
    class VERT:
        co=CO()
    
    if v2==None: #CHECKS FOR A VERTEX.
        #returns 1 if v lies within 'loop'. returns 0 if v is part of the loop 'loop'. return -1 if v lies outside 'loop'.
        #ALGO:: When the max. no. of edges intersected by the line drawn from the given pt. to a pt. lying on the loop is
        #       even (including zero) then this pt. lies within the loop, else outside it.
        
        v=v1
        v1=vert[v]
        i=0
        floop=[] #fake loop
        for vi in loop:
            if not ignoreThatVertexLieOnTheLoop:
                if (vert[vi].co.x,vert[vi].co.y)==(v1.co.x,v1.co.y):  return 0
            
            if vert[vi].co.x!=v1.co.x:
                floop.append(vi)
            
            if i==0: ymax=vert[vi].co.y
            if vert[vi].co.y>ymax:
                ymax=vert[vi].co.y
            i+=1
        
        v2=VERT()
        (v2.co.x,v2.co.y)=(v1.co.x,ymax+0.01)
        
        loop=floop
        edgeCount=0
        for i in range(len(loop)):
            vi3=loop[i]
            v3=vert[vi3]
            if i==(len(loop)-1):
                vi4=loop[0]
            else:
                vi4=loop[i+1]
            v4=vert[vi4]
            f=isLinesIntersecting(v1,v2,v3,v4,True)
            if f==2 or f==3 or f==4:
                if DEBUG: print 'Invalid return. Expected 0 or 1. Got '+str(f)
                return
            if f==1: edgeCount+=1
        if edgeCount%2==0:
            return -1
        else:
            return 1
        
    else:# CHECKS FOR AN EDGE.
        #return True if edge v1-v2 lie in loop 'loop' else false, where v1 & v2 lie on the given loop.
        #ALGO:: An edge is inside a loop if the mid-pt. of this line segment (edge) is within the loop
        #       else it is outside. Actually any pt. on this line segment can be taken but for simplicity
        #       sake I have chosen the mid-pt.
        i1=loop.index(v1)
        i2=loop.index(v2)
        #the given edge is along the loop's edge.
        if (i1==len(loop)-1 and i2==0) or (i2==len(loop)-1 and i1==0): return True
        if i1==i2+1 or i1+1==i2: return True
        
        midv=VERT()
        (midv.co.x,midv.co.y)=((vert[v1].co.x+vert[v2].co.x)/2,(vert[v1].co.y+vert[v2].co.y)/2)
        
        fVert=vert[0:] #making fake vert list for the purpose
        fVert.append(midv)
        
        f=isInLoop(loop=loop,vert=fVert,v1=len(fVert)-1,ignoreThatVertexLieOnTheLoop=True)
        
        if f==-1:
            return False
        else:
            return True
    
def bubblesort(l,by,start,end): #bubble sorts the portion of list l starting from 'start' to 'end'(exclusive)
    #according to increasing values of 'by'.
    for i in range(start,end):
        DrawProgressBar(i/end,"Bubble Sorting")
        for j in range(i+1,end):
            if l[i][by]>l[j][by]:
                t=l[i]
                l[i]=l[j]
                l[j]=t
    return l

def MOD(n):
    if n<0:
        return -1*n
    else:
        return n

def vert2face(makeQuads,detectLoops):
    selection = Object.GetSelected()
    if len(selection) == 0:
        error('No object selected')
        return
        
    if len(selection)>1:
        error('Select only one object')
        return
    
    if selection[0].getType() != 'Mesh':
        error('Only mesh are allowed.')
        return
    
    obj=selection[0].getData() #getting the object.
    
    #getting list of all selected vertices.
    vert=[]
    first=True
    for v in obj.verts:
        if first:
            first=False
            minZ=v.co.z
            maxZ=v.co.z
        else:
            if v.co.z<minZ: minZ=v.co.z
            if v.co.z>maxZ: maxZ=v.co.z
        if v.sel:
            vert.append(v)
    
    #getting list of edges, even if none of the vertices of it are selected.
    Edges=[]
    for e in obj.edges:
        v1=e.v1
        v2=e.v2
        if v1.sel and v2.sel: #we don't want the selected edges again in the list of edges to be checked for crossing.
            continue
        v1=v1.co
        v2=v2.co
        if (v1.z<minZ and v2.z<minZ) or (v1.z>maxZ and v2.z>maxZ):
            continue
        Edges.append(e)

    #getting selected faces and the vertices which are not to be connected.
    face=[]
    doNotConnect=[]
    for f in obj.faces:
        selVCount=0
        totVCount=0
        selv=[]
        for v in f.v:
            totVCount+=1
            if v.sel:
                selVCount+=1
                selv.append(v)
        if selVCount==4: face.append(f)
        if totVCount==3 and selVCount==3:
            face.append(f)
            continue
        if selVCount==3 and totVCount==4:
            for i in range(len(selv)):
                for j in range(i+1,len(selv)):
                    if obj.findEdge(selv[i],selv[j]) is None:
                        doNotConnect.append([selv[i],selv[j]])

    #removing selected faces.
    for f in face:
        edge=[]
        for i in range(len(f.v)):
            for j in range(i+1,len(f.v)):
                e=obj.findEdge(f.v[i],f.v[j])
                if e is not None:
                    edge.append(e)
        #now removing the face.
        obj.removeFace(f)
        #recreating removed edges. (Vertices are not automatically deleted. Hence no fuss with that)
        for e in edge:
            obj.addEdge(e.v1,e.v2)
    obj.update()
    
    ##Hence for revision.##
    #obj is the mesh object.
    #vert is the list of selected vertices.
    #doNotConnect is a list of vertices NOT to be connected.
    
    if DEBUG:
        print "==Vertices data dump=="
        i=0
        for v in vert:
            print "vert "+str(i)+" ("+str(v.co.x)+","+str(v.co.y)+") ::-> "+str(v)
            i=i+1
        print("==Dump complete==")
    
    #calculating distances of each vertice with every other.
    stdDev=[]
    for i in range(len(vert)):
        DrawProgressBar(i/len(vert),'Calculating dis. of each vert with every other')
        r=[]
        for j in range(len(vert)):
            if i==j: continue
            dis=calDis(vert[i],vert[j])
            if dis==0:
                error('Maybe faulty mesh. Two vertices in the same location.')
                return
            r.append(dis)
        stdDev.append([calStdDeviation(r),i])
    
    #bubble sorting stdDev according to increasing values of std deviation.
    stdDev=bubblesort(stdDev,0,0,len(stdDev))
    
    if DEBUG:
        print '@@stdDev dump@@'
        for s in stdDev:
            print s
        print '@@Dump complete@@'
    
    #sorting vert according to increasing stdDev.
    tvert=[]
    P=0
    for sd in stdDev:
        tvert.append(vert[sd[1]])
        DrawProgressBar(P/len(stdDev),'Sorting vertices according to stdDev')
        P+=1
    vert=tvert
    tvert=[]#clearing it
    #stdVert=[]#clearing it
    
    #Making distance matrix.
    if USE_EXPERIMENTAL_ALGO:
        disMat=[]
        for i in range(len(vert)):
            DrawProgressBar(i/len(vert),'Calculating dis. of each vert with every other')
            r=[]
            for j in range(len(vert)):
                if i==j: continue
                dis=calDis(vert[i],vert[j])
                if dis==0:
                    error('Maybe faulty mesh. Two vertices in the same location.')
                    return
                r.append([dis,j])
                #loc.append(j)
            #R=[r,loc]
            #R=bubblesort(R,0,0,len(R))
            r=bubblesort(r,0,0,len(r))
            R=[]
            for l in r:
                R.append(l[1])
            disMat.append(R)
##        if DEBUG:
##            print '()()()disMat dump()()()'
##            for d in disMat:
##                print d
##            print '()()()dump complete()()()'
    
    vd=[] #stores index of all the vertices a particular vertex is connected to.
    #first searching and updating vd of already connected vertices.
    for i in range(len(vert)):
        DrawProgressBar(i/len(vert),'finding already connected verts')
        r=[]
        for j in range(i+1,len(vert)):
            if obj.findEdge(vert[i],vert[j]) is not None:
                r.append(j)
        vd.append(r)
##    if DEBUG:
##        print '**vd dump (of already connected vertices)**'
##        for v in vd:
##            print v
##        print '**Dump complete**'
    
    #Detecting loops.
    outerMostLoop=[]
    if detectLoops:
        loops=DetectLoops(vd)
        if len(loops)!=0:
            rank=RankLoops(loops,vert)
            Loops=[]
            for i in range(len(loops)):
                t=[rank[i]]
                t.extend(loops[i])
                Loops.append(t)
            if DEBUG: print '::::::rank::::::: '+str(rank)
            rank=[]
            
            Loops=bubblesort(Loops,0,0,len(Loops))
            
            if DEBUG:
                print '[o][o][o]Loops dump (sorted)[o][o][o]'
                for loo in Loops:
                    print loo
                print '[o][o][o]dump complete[o][o][o]'
            
            #Making Loops[0] (the outermost loop) linkable to itself only from inside, not outside.
            outerMostLoop=Loops[0][1:]
##            lo=Loops[0][1:]
##            print lo
##            for i in range(len(lo)):
##                for j in range(len(lo)):
##                    if i==j: continue
##                    if i==len(lo)-1:
##                        k=0
##                    else:
##                        k=i+1
##                    if lo[k]==lo[j]: continue
##                    if i==0:
##                        k=len(lo)-1
##                    else:
##                        k=i-1
##                    if lo[k]==lo[j]: continue
##                    f=isInLoop(lo,vert,lo[i],lo[j])
##                    print str((lo[i],lo[j])),f
##                    if not f:
##                        doNotConnect.append([vert[lo[i]],vert[lo[j]]])
            
            #Now detecting holes.
            if Loops[0][0]%2==0: #rank of loop 0.
                Factor=0
            else:
                Factor=1
            loops=[]
            for i in range(1,len(Loops)):
                if (Loops[i][0]+Factor)%2==0: continue
                
                thisIsLastLoop=True
                if Loops[i][0]==Loops[i-1][0]:
                    td=[]
                    for t in Loops[i][1:]:
                        td.append(vert[t])
                    if DEBUG: print '0>loop '+str(i)+' donotconnect vert '+str(Loops[i][1:])
                    doNotConnect[len(doNotConnect)-1].extend(td)
                    continue
                
                for j in range(i+1,len(Loops)):
                    if Loops[i][0]==(Loops[j][0]-1): #rank of current == rank of jth loop-1
                        k=Loops[j][0] #chossing an arbitrary vertex (here zero) to test that loop j is in loop i & not in any other loop of same rank as that of i.
                        if isInLoop(Loops[i][1:],vert,k):
                            thisIsLastLoop=False
                            for t2 in Loops[j][1:]:
                                t=Loops[i][1:]
                                t.append(t2)
                                td=[]
                                for t1 in t:
                                    td.append(vert[t1])
                                if DEBUG: print '1>loop '+str(i)+' donotconnect vert '+str(t)
                                doNotConnect.append(td)
                
                if thisIsLastLoop:
                    td=[]
                    for t in Loops[i][1:]:
                        td.append(vert[t])
                    if DEBUG: print '2>loop '+str(i)+' donotconnect vert '+str(Loops[i][1:])
                    doNotConnect.append(td)
                    if len(Loops[i])==4:
                        loops.append(Loops[i][1:]) #This loops will be used while making vertexList for triangles. Since if 3 verts for a loop then that wiil be filled.
            Loops=[]
##    if DEBUG:
##        print '|||||||doNotConnect dump||||||||'
##        for d in doNotConnect:
##            print d
##        print '|||||||dump complete||||||||||||'
    
    ##now connecting vertices.
    if USE_EXPERIMENTAL_ALGO:
        I=0
        while I <len(vert):
            DrawProgressBar(I/len(vert),'Connecting verts')
            for j in range(I,len(stdDev)):
                if stdDev[I][0]==stdDev[j][0]:
                    J=j
                else:
                    break
            
            for j in range(len(vert)-1):
                for i in range(I,J+1):
                    k=disMat[i][j]
                    if k<i: continue
                    if vd[i].count(k)==0 and isConnectable(vert,vd,doNotConnect,Edges,i,k):
                        if detectLoops and outerMostLoop!=[]:
                            if outerMostLoop.count(i)!=0 and outerMostLoop.count(k)!=0:
                                if isInLoop(outerMostLoop,vert,i,k):
                                    vd[i].append(k)
                            else:
                                vd[i].append(k)
                        else:
                            vd[i].append(k)
            I=J+1
    else:
        #this is the old algo.
        for i in range(len(vert)):
            DrawProgressBar(i/len(vert),'Connecting verts')
            for j in range(i+1,len(vert)):
                if vd[i].count(j)==0 and isConnectable(vert,vd,doNotConnect,Edges,i,j):
                    if detectLoops and outerMostLoop!=[]:
                        if outerMostLoop.count(i)!=0 and outerMostLoop.count(j)!=0:
                            if isInLoop(outerMostLoop,vert,i,j):
                                vd[i].append(j)
                        else:
                            vd[i].append(j)
                    else:
                        vd[i].append(j)
    stdDev=[]#clearing stdDev
    
##    if DEBUG:
##        print '**vd dump (Final)**'
##        for v in vd:
##            print v
##        print '**Dump complete**'
    
    #making faces (tris). :-)
    vertexList=[]
##    if DEBUG: print '000vertexList dump000'
    for vi in range(len(vd)):
        DrawProgressBar(vi/len(vd),'Making faces :) (tris)')
        vil=vd[vi]
        for i in vil:
            for j in vil:
                if i==j: continue
                goto_outer=False
                if detectLoops:
                    for l in loops:
                        if l.count(vi)!=0 and l.count(i)!=0 and l.count(j)!=0:
                            goto_outer=True
                            break
                            #because these 3 vertices are part of loops & hence should not be filled.
                if goto_outer: continue
                if vd[i].count(j)!=0: #means vert[i] is connected with vert[j].
                    vertices=[vert[vi],vert[i],vert[j]]
                    if isTrigulizable(vertices,vert):
                        vertexList.append(vertices)
##                        if DEBUG: print [vi,i,j]
##    if DEBUG: print '000dump complete000'


    ########Part 2. Converting tris to quads.#########
    if makeQuads:
        #scannig for tris with common vertices.
        triPair=[]
        for i in range(len(vertexList)):
            DrawProgressBar(i/len(vertexList),'Scanning for mergeable tris')
            for j in range(i+1,len(vertexList)):
                commonCount=0
                singleVert=[]
                commonVert=[]
                v=[]
                V1=[]
                V1.extend(vertexList[i])
                V2=[]
                V2.extend(vertexList[j])
                for vi in vertexList[i]:
                    for vj in vertexList[j]:
                        if (vi.co.x,vi.co.y)==(vj.co.x,vj.co.y):
                            commonVert.append(vi)
                            V1.remove(vi)
                            V2.remove(vj)
                            commonCount+=1
                            break
                singleVert.extend(V1)
                singleVert.extend(V2)
                v.append(singleVert)
                v.append(commonVert)
                
##                if DEBUG:
##                    if commonCount!=2: print 'ooo> Tris '+str(i)+' & '+str(j)+' unpaired, since common vertices count!=2'
                if commonCount==2:
                    tv=[]
                    tv.extend(v[0])
                    tv.extend(v[1])
                    if not isInSamePlane(tv): continue #vertices not in same plane hence shouldn't be connected.
                    vec=[]
                    for I in range(len(v)):
                        for iI in range(len(v[I])):
                            v1=v[I][iI]
                            if I==0: J=1
                            else: J=0
                            for v2 in v[J]:
                                ptx=v2.co.x-v1.co.x
                                pty=v2.co.y-v1.co.y
                                vec.append(Mathutils.Vector([ptx,pty]))
                            if iI==0: iJ=1
                            else: iJ=0
                            v2=v[I][iJ]
                            ptx=v2.co.x-v1.co.x
                            pty=v2.co.y-v1.co.y
                            vec.append(Mathutils.Vector([ptx,pty]))
                    
                    SKIP=False
                    ang=[]
                    dis=[]
                    for k in range(0,len(vec),3):
                        angle1=Mathutils.AngleBetweenVecs(vec[k],vec[k+2])
                        angle2=Mathutils.AngleBetweenVecs(vec[k+1],vec[k+2])
                        angle=angle1+angle2
                        if angle>=180: #when ANY of the angles in a quad is 180 or more then it is better to have tris than such a quad.
                            SKIP=True
                            break
                        ang.append(angle)
                        dis.append(vec[k].length);
                    if SKIP:
##                        if DEBUG: print 'ooo> Tris '+str(i)+' & '+str(j)+' unpaired, since angle>=180'
                        continue
                    else:
                        triPair.append([i,j,calStdDeviation(ang,90),calStdDeviation(dis)])
##        if DEBUG:
##            print '""""triPair dump(unsorted)""""'
##            print triPair
##            print '""""dump complete""""'
        
        #sorting triPair according to increasing values of angle deviation.
        triPair=bubblesort(triPair,2,0,len(triPair))
        #distance deviation is used for sorting pairs with same angle deviation.
        for i in range(len(triPair)):
            sameCount=0
            for j in range(i+1,len(triPair)):
                if triPair[j][2]!=triPair[i][2]: break
                sameCount+=1
            if sameCount==0: continue
            triPair=bubblesort(triPair,3,i,i+sameCount+1)
            i+=sameCount
        
        #removing all pairs in triPair that have any of their vertices present already in the pairs preceeding it.
        for i in range(len(triPair)):
            if triPair[i]==[]: continue
            for j in range(i+1,len(triPair)):
                if triPair[j]==[]: continue
                if triPair[i][0]==triPair[j][0] or triPair[i][0]==triPair[j][1] or triPair[i][1]==triPair[j][0] or triPair[i][1]==triPair[j][1]:
                    triPair[j]=[] #actually they are not removed but are set to [].
##        if DEBUG:
##            print
##            print '""""triPair dump(final)""""'
##            print triPair
##            print '""""dump complete""""'
        
        #now making quad faces.
        face=[]
        P=0
        for t in triPair:
            DrawProgressBar(P/len(triPair),'Making quads')
            P+=1
            if t==[]: continue
            v=[]
            v.extend(vertexList[t[0]])
            #searching and appending only the vertices which are not in v.
            for vrt in vertexList[t[1]]:
                present=False
                for vrt1 in vertexList[t[0]]:
                    if (vrt1.co.x,vrt1.co.y)==(vrt.co.x,vrt.co.y):
                        v.remove(vrt1)
                        v.append(vrt1)
                        present=True
                if not present:
                    alone=vrt
            obj.removeEdge(v[1],v[2]) #removing the common edge from the mesh.
            v.insert(2,alone)
            vertexList[t[0]]=[]
            vertexList[t[1]]=[]
##            if DEBUG:
##                print 'final face vertex list->'+str(v)
            face.append(NMesh.Face(v))

    #now making tris faces.
    if not makeQuads: face=[]
    for t in vertexList:
        if t==[]: continue
        face.append(NMesh.Face(t))

    #updating obj.
    for f in face:
        obj.addFace(f)
    
    obj.update()

##calling vert2face##
Window.DrawProgressBar(0.0, 'Vertices to Faces v '+__version__)
print '========================================'
print 'Vertices to Faces v '+__version__+'. Processing...'
result=Draw.PupMenu("Make quads?%t|Yes|No")
if result==-1 or result==2:
    makeQuads=False
else:
    makeQuads=True

result=Draw.PupMenu("Detect loops?%t|Yes|No")
if result==-1 or result==2:
    detectLoops=False
else:
    detectLoops=True

is_editmode = Window.EditMode()
if is_editmode: Window.EditMode(0)
vert2face(makeQuads,detectLoops)
if is_editmode: Window.EditMode(1)
print 'Done!'
print '========================================'
Window.DrawProgressBar(1.0, '')
